import pandas as pd

def load_data(path):
    """
    Load csv data
    :param path: path to csv data
    :return: dataframe
    """
    data = pd.read_csv(path)
    return data


# Analysis of punctuality
def mean_time_difference(bus_data_with_punctuality_path):
    """
    Calculate mean time difference in minutes between timetible arrival and actual bus arrival
    :param bus_data_with_punctuality_path: path to file containing bus data with punctuality
    :return: mean time difference in minutes
    """
    bus_data_with_punctuality_df = load_data(bus_data_with_punctuality_path)
    mean_time_diff = bus_data_with_punctuality_df['TimeDiff'].mean()
    mean_time_diff = round(mean_time_diff, 2)
    return mean_time_diff


def max_time_difference(bus_data_with_punctuality_path):
    """
    Find bus with maximal time difference in minutes between timetible arrival and actual bus arrival
    :param bus_data_with_punctuality_path: path to file containing bus data with punctuality
    :return: information about bus which has maximal time difference in minutes
    """
    bus_data_with_punctuality_df = load_data(bus_data_with_punctuality_path)
    max_time_diff_ind = bus_data_with_punctuality_df['TimeDiff'].idxmax() # index with highest vale in column
    max_time_diff_bus = bus_data_with_punctuality_df.loc[[max_time_diff_ind]]
    return max_time_diff_bus


def find_buses_exceeding_speed_threshold(buses_velocity_path, speed_threshold): # velocity in km/h
    """
    Find buses which exceeded given speed threshold
    :param buses_velocity_path: path to file containing bus data velocity
    :param speed_threshold: int, speed threshold in kmph
    :return: informations about buses exceeding speed threshold in kmph
    """
    buses_velocity_df = load_data(buses_velocity_path)
    buses_exceeding_speed = buses_velocity_df.loc[buses_velocity_df['Velocity'] > int(speed_threshold)] # to be sure int
    return buses_exceeding_speed



import requests
import pandas as pd
import time
import datetime


def get_buses_data(frequence, timerange, path): # both in min
    """
    Download Autobusy i tranwaje online from Otwarte dane po warszawsku
    :param frequence: frequency of requesting data, in minutes
    :param timerange: duration of collecting data, in minutes
    :param path: path to save collected data
    """
    start_datetime = datetime.datetime.now()
    print(start_datetime, "Dowloading bus data...")
    span_added = datetime.timedelta(minutes=timerange)
    end_date_and_time = start_datetime + span_added

    while datetime.datetime.now() <= end_date_and_time:
        try:
            response = requests.get(
            "https://api.um.warszawa.pl/api/action/busestrams_get/?resource_id= f2e5503e927d-4ad3-9500-4ab9e55deb59&apikey=a0069121-f1a7-4b53-9c37-855f3e75748e&type=1"
            )
            df = pd.json_normalize(response.json(), 'result')
            df.to_csv(f'{path}.csv', mode='w+', header=False, index=False)
            time.sleep(frequence * 60)
        except TypeError:
            pass
    print(datetime.datetime.now(), "Completed downloading bus data")


def reformat_json(raw_json):
    """
    Reformat api response to list of python dictionaries
    :param raw_json: response from the api
    :param reformatted: list of python dictionaries
    """
    reformatted = []
    for dic in raw_json:
        values = dic['values']
        details_dict = {}
        for value in values:
            details_dict[value['key']] = value['value']
        reformatted.append(details_dict)
    return reformatted


def remove_stops_with_nulls(reformatted):
    """
    Remove stops with null coordinates
    :param reformatted: reformatted api reponse
    :return: nulls_removed: only stops with not null coordinates
    """
    nulls_removed = []
    for dic in reformatted:
        if dic['szer_geo'] != 'null' or dic['dlug_geo'] != 'null':
            nulls_removed.append(dic)
    return nulls_removed


def get_stops_data(path):
    """
    Download Współrzędne przystanków from Otwarte dane po warszawsku
    :param path: path to save collected data
    """
    response = requests.get(
        'https://api.um.warszawa.pl/api/action/dbstore_get?id=ab75c33d-3a26-4342-b36a-6e5fef0a3ac3&apikey=a0069121-f1a7-4b53-9c37-855f3e75748e'
    )

    result = response.json()['result']
    reformatted_result = reformat_json(result)
    nulls_removed_result = remove_stops_with_nulls(reformatted_result)
    df = pd.DataFrame(nulls_removed_result)
    print(df)
    df.to_csv(f'{path}.csv', mode='w', header=False, index=False)
    print(datetime.datetime.now(), "Completed downloading stops data")


def collect_bus_and_stops_data(frequence, timerange, bus_data_path, stops_data_path):
    get_buses_data(frequence, timerange, bus_data_path)
    get_stops_data(stops_data_path)


def get_timetable_for_line(stop_id, stop_nr, line):
    """
    Download Rozkłady jazdy ZTM from Otwarte dane po warszawsku
    :param stop_id: id of stop
    :param stop_nr: nr of stop
    :param line: bus line
    """
    stop_nr = str(stop_nr)
    if len(stop_nr) == 1:
        stop_nr = '0' + stop_nr

    parameters = {
        "busstopId": str(stop_id),
        "busstopNr": stop_nr,
        "line": str(line)
    }
    response = requests.get('https://api.um.warszawa.pl/api/action/dbtimetable_get?id=e923fa0e-d96c-43f9-ae6e-60518c9f3238&apikey=a0069121-f1a7-4b53-9c37-855f3e75748e',
                            params=parameters
                            )

    result = response.json()['result']
    reformatted_result = reformat_json(result)
    return reformatted_result



import datetime
import re
import sys
import pandas as pd
from geopy import distance
from wawpybus.collect import get_timetable_for_line


def load_bus_data(path):
    """
    Load file with bus data
    :param path: path to bus data file
    :return: dataframe
    """
    data = pd.read_csv(f'{path}', names=['Line', 'Lon', 'VehicleNumber', 'Datetime', 'Lat', 'Brigade'])
    data = data.sort_values(by=['Line', 'Brigade'], ignore_index=True)
    data['VehicleCoords'] = list(zip(data.Lat, data.Lon))
    return data


def load_stops_data(path):
    """
    Load file with stops data
    :param path: path to stops data file
    :return: dataframe
    """
    data = pd.read_csv(path,
                       names=['StopID', 'StopNumber', 'StopName', 'StreetID', 'Lat', 'Lon', 'Direction', 'ValidFrom'])
    data = data.sort_values(by=['StopID', 'StopNumber'], ignore_index=True)
    data['StopCoords'] = list(zip(data.Lat, data.Lon))
    return data


def process_measurement(measurement, measurements_df, stops_data):
    vehicle_coords = measurement['VehicleCoords']
    time = measurement['Datetime'].split()[1]  # separate date and time and select time
    min_dist = sys.maxsize
    nearest_stop = None
    for ind, stop in stops_data.iterrows():
        stop_coords = stop['StopCoords']
        dist = distance.great_circle(vehicle_coords, stop_coords).m
        if dist < min_dist:
            min_dist = dist
            nearest_stop = stop
    measurement_dict = {'VehicleNumber': measurement['VehicleNumber'], 'VehicleCoords': measurement['VehicleCoords'],
                        'Line': measurement['Line'], 'Time': time, 'StopID': nearest_stop['StopID'],
                        'StopNumber': nearest_stop['StopNumber'], 'StopName': nearest_stop['StopName'],
                        'Dist': round(min_dist, 2)}
    keys_values = measurement_dict.items()
    measurement_dict = {key: str(value) for key, value in keys_values}
    measurements_df = measurements_df.append(measurement_dict, ignore_index=True)
    return measurements_df


def find_closest_stops(bus_data_path, stops_data_path, bus_data_with_nearest_stops_path, lines: list = None):
    """
    Find stops closest to given bus position by coordinates
    :param bus_data_path: path to bus data file
    :param stops_data_path: path to stops data file
    :param bus_data_with_nearest_stops_path: path to output file
    :param lines: list of bus line to analyse, if not given analyse all bus lines included in bus data
    """
    stops_data = load_stops_data(stops_data_path)
    bus_data = load_bus_data(bus_data_path)

    if lines is None:
        lines_l = bus_data.Line.unique()  # if parameter lines is not given check all lines
    else:
        lines_l = lines
    bus_lines = bus_data.loc[bus_data['Line'].isin(lines_l)]  # bus_lines is a subset of bus_data

    bus_data_with_nearest_stops_df = pd.DataFrame()
    for ind, measurement in bus_lines.iterrows():
        bus_data_with_nearest_stops_df = process_measurement(measurement, bus_data_with_nearest_stops_df, stops_data)

    bus_data_with_nearest_stops_df = bus_data_with_nearest_stops_df.sort_values(by=['VehicleNumber', 'Time'],
                                                                                ignore_index=True)
    bus_data_with_nearest_stops_df.to_csv(f'{bus_data_with_nearest_stops_path}.csv', header=True, index=False)


def process_timetable_item(measurement, minimal_diff, nearest_time_from_timetable, timetable_item):
    timetable_item_time = re.sub('^(24)', '00', timetable_item['czas'])  # in data h 00 was written as 24
    timetable_dtobj = datetime.datetime.strptime(timetable_item_time, '%H:%M:%S')
    time_dtobj = datetime.datetime.strptime(measurement['Time'], '%H:%M:%S')
    if timetable_dtobj > time_dtobj:
        diff = timetable_dtobj - time_dtobj
    else:
        diff = time_dtobj - timetable_dtobj
    diff = diff.seconds / 60
    if diff < minimal_diff:
        minimal_diff = diff  # in min
        nearest_time_from_timetable = timetable_dtobj
    return minimal_diff, nearest_time_from_timetable


def check_punctuality(bus_data_with_nearest_stops_path, bus_data_with_punctuality_path):
    """
    Calculate difference between bus time arrival at stop and time given in timetable
    :param bus_data_with_nearest_stops_path: path to bus data with nearest stops file
    :param bus_data_with_punctuality_path: path to output data
    """
    measurements = pd.read_csv(bus_data_with_nearest_stops_path)

    bus_data_with_stops_df = pd.DataFrame()
    for ind, measurement in measurements.iterrows():
        timetable = get_timetable_for_line(measurement['StopID'], measurement['StopNumber'], measurement['Line'])

        if timetable:  # means bus stops at this stop
            minimal_diff = sys.maxsize
            nearest_time_from_timetable = None

            for timetable_item in timetable:
                minimal_diff, nearest_time_from_timetable = process_timetable_item(measurement, minimal_diff,
                                                                                   nearest_time_from_timetable,
                                                                                   timetable_item)

            bus_data_with_stops = {'TimeDiff': round(minimal_diff, 2), 'VehicleNumber': measurement['VehicleNumber'],
                                   'Line': measurement['Line'], 'Direction': timetable_item['kierunek'],
                                   'StopID': measurement['StopID'], 'StopNumber': measurement['StopNumber'],
                                   'Time': measurement['Time'], 'StopName': measurement['StopName'],
                                   'NearestTimeFromTimetable': nearest_time_from_timetable}
            keys_values = bus_data_with_stops.items()
            bus_data_with_stops = {key: str(value) for key, value in keys_values}
            bus_data_with_stops_df = bus_data_with_stops_df.append(bus_data_with_stops, ignore_index=True)

    bus_data_with_stops_df.to_csv(f'{bus_data_with_punctuality_path}.csv', header=True, index=False)


def check_if_exceeded_speed(bus_data_with_nearest_stops_path, buses_velocity_path):
    """
    Calculate bus speed between consecutive positions in kmph
    :param bus_data_with_nearest_stops_path: path to bus data with nearest stops file
    :param buses_velocity_path: path to bud velocity data file
    """
    bus_data_with_nearest_stops = pd.read_csv(bus_data_with_nearest_stops_path)
    vehicles = bus_data_with_nearest_stops.VehicleNumber.unique()

    buses_velocity_df = pd.DataFrame()
    for vehicle in vehicles:
        vehicle_measurements = bus_data_with_nearest_stops[bus_data_with_nearest_stops.VehicleNumber == vehicle]
        previous = vehicle_measurements.iloc[0]
        for ind, measurement in vehicle_measurements.iterrows():
            if ind == 0:
                continue
            current = measurement

            dist = distance.great_circle(eval(previous['VehicleCoords']), eval(current['VehicleCoords'])).km

            previous_dtobj = datetime.datetime.strptime(previous['Time'], '%H:%M:%S')
            current_dtobj = datetime.datetime.strptime(current['Time'], '%H:%M:%S')

            diff = current_dtobj - previous_dtobj
            diff = diff.seconds / 3600  # to obtain an hour

            velocity = dist / diff if diff != 0 else dist
            bus_velocity = {'VehicleNumber': measurement['VehicleNumber'], 'Line': measurement['Line'],
                            'EndPosition': current['VehicleCoords'], 'StartPosition': previous['VehicleCoords'],
                            'EndTime': current['Time'], 'StartTime': previous['Time'],
                            'EndNearestStopName': current['StopName'], 'StartNearestStopName': previous['StopName'],
                            'EndNearestStopNr': current['StopNumber'], 'StartNearestStopNr': previous['StopNumber'],
                            'Velocity': round(velocity, 2)}
            keys_values = bus_velocity.items()
            bus_velocity = {key: str(value) for key, value in keys_values}
            buses_velocity_df = buses_velocity_df.append(bus_velocity, ignore_index=True)

            previous = current

    buses_velocity_df.to_csv(f'{buses_velocity_path}.csv', header=True, index=False)
